/*
 * Copyright (c) 1993 Christopher G. Demetriou
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Christopher G. Demetriou.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software withough specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: exec_aout.c,v 1.1.2.5 1993/10/17 04:29:26 deraadt Exp $
 */

#include "param.h"
#include "systm.h"
#include "filedesc.h"
#include "kernel.h"
#include "proc.h"
#include "mount.h"
#include "malloc.h"
#include "namei.h"
#include "vnode.h"
#include "file.h"
#include "exec.h"
#include "resourcevar.h"
#include "wait.h"

#include "machine/cpu.h"
#include "machine/reg.h"
#include "machine/exec.h"

#include "mman.h"
#include "vm/vm.h"
#include "vm/vm_param.h"
#include "vm/vm_map.h"
#include "vm/vm_kern.h"
#include "vm/vm_pager.h"

/* define some macros usually in <a.out.h> that come in handy later */
#define N_TXTADDR(x,m) __LDPGSZ
#define N_DATADDR(x,m) (((m)==OMAGIC) ? (N_TXTADDR(x,m) + (x).a_text) \
			: (__LDPGSZ + ((N_TXTADDR(x,m) + (x).a_text - 1) \
				       & ~(__LDPGSZ-1))))
#define N_BSSADDR(x,m) (N_DATADDR(x,m)+(x).a_data)

#define N_TXTOFF(x,m) ((m)==ZMAGIC ? 0 : sizeof (struct exec))
#define N_DATOFF(x,m) (N_TXTOFF(x,m) + (x).a_text)

/*
 * exec_aout_makecmds(): Check if it's an a.out-format executable.
 *
 * Given a proc pointer and an exec package pointer, see if the referent
 * of the epp is in a.out format.  First check 'standard' magic numbers for
 * this architecture.  If that fails, try a cpu-dependent hook.
 *
 * This function, in the former case, or the hook, in the latter, is
 * responsible for creating a set of vmcmds which can be used to build
 * the process's vm space and inserting them into the exec package.
 */

int
exec_aout_makecmds(p, epp)
	struct proc *p;
	struct exec_package *epp;
{
	u_long midmag, magic;
	u_short mid;
	int error;

	midmag = ntohl(epp->ep_execp->a_midmag);
	mid = (midmag >> 16) & 0x3ff;
	magic = midmag & 0xffff;

#ifdef EXEC_DEBUG
	printf("exec_makecmds: a_midmag is %x, magic=%x mid=%x\n",
	    epp->ep_execp->a_midmag, magic, mid);
#endif

	midmag = mid << 16 | magic;

	switch (midmag) {
	case (MID_MACHINE << 16) | ZMAGIC:
		error = exec_aout_prep_zmagic(p, epp);
		break;
	case (MID_MACHINE << 16) | NMAGIC:
		error = exec_aout_prep_nmagic(p, epp);
		break;
	case (MID_MACHINE << 16) | OMAGIC:
		error = exec_aout_prep_omagic(p, epp);
		break;
	default:
		error = cpu_exec_aout_makecmds(p, epp);
	}

	if (error && epp->ep_vcp)
		kill_vmcmd(&epp->ep_vcp);

bad:

#ifdef EXEC_DEBUG
	printf("exec_makecmds returning with error = %d\n", error);
#endif
	return error;
}

/*
 * exec_aout_prep_zmagic(): Prepare a 'native' ZMAGIC binary's exec package
 *
 * First, set of the various offsets/lengths in the exec package.
 *
 * Then, mark the text image busy (so it can be demand paged) or error
 * out if this is not possible.  Finally, set up vmcmds for the
 * text, data, bss, and stack segments.
 */

int
exec_aout_prep_zmagic(p, epp)
	struct proc *p;
	struct exec_package *epp;
{
	struct exec *execp = epp->ep_execp;
	struct exec_vmcmd *ccmdp;

	epp->ep_taddr = N_TXTADDR (*execp, ZMAGIC);
	epp->ep_tsize = execp->a_text;
	epp->ep_daddr = N_DATADDR (*execp, ZMAGIC);
	epp->ep_dsize = execp->a_data + execp->a_bss;
	epp->ep_entry = execp->a_entry;

	/*
	 * check if vnode is in open for writing, because we want to
	 * demand-page out of it.  if it is, don't do it, for various
	 * reasons
	 */
	if ((execp->a_text != 0 || execp->a_data != 0) &&
	    (epp->ep_vp->v_flag & VTEXT) == 0 && epp->ep_vp->v_writecount != 0) {
#ifdef DIAGNOSTIC
		if (epp->ep_vp->v_flag & VTEXT)
			panic("exec: a VTEXT vnode has writecount != 0\n");
#endif
		epp->ep_vcp = NULL;
		return ETXTBSY;
	}
	epp->ep_vp->v_flag |= VTEXT;

	/* set up command for text segment */
	epp->ep_vcp = new_vmcmd(vmcmd_map_pagedvn,
	    execp->a_text,
	    epp->ep_taddr,
	    epp->ep_vp,
	    N_TXTOFF(*execp, ZMAGIC),
	    VM_PROT_READ | VM_PROT_EXECUTE);
	ccmdp = epp->ep_vcp;

	/* set up command for data segment */
	ccmdp->ev_next = new_vmcmd(vmcmd_map_pagedvn,
	    execp->a_data,
	    epp->ep_daddr,
	    epp->ep_vp,
	    N_DATOFF(*execp, ZMAGIC),
	    VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE);
	ccmdp = ccmdp->ev_next;

	/* set up command for bss segment */
	ccmdp->ev_next = new_vmcmd(vmcmd_map_zero,
	    execp->a_bss,
	    epp->ep_daddr + execp->a_data,
	    0,
	    0,
	    VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE);
	ccmdp = ccmdp->ev_next;

	return exec_aout_setup_stack(p, epp, ccmdp);
}

/*
 * exec_aout_prep_nmagic(): Prepare a 'native' NMAGIC binary's exec package
 */

int
exec_aout_prep_nmagic(p, epp)
	struct proc *p;
	struct exec_package *epp;
{
	struct exec *execp = epp->ep_execp;
	struct exec_vmcmd *ccmdp;
	long bsize, baddr;

	epp->ep_taddr = N_TXTADDR(*execp, NMAGIC);
	epp->ep_tsize = execp->a_text;
	epp->ep_daddr = N_DATADDR(*execp, NMAGIC);
	epp->ep_dsize = execp->a_data + execp->a_bss;
	epp->ep_entry = execp->a_entry;

	/* set up command for text segment */
	epp->ep_vcp = new_vmcmd(vmcmd_map_readvn,
	    execp->a_text,
	    epp->ep_taddr,
	    epp->ep_vp,
	    N_TXTOFF(*execp, NMAGIC),
	    VM_PROT_READ | VM_PROT_EXECUTE);
	ccmdp = epp->ep_vcp;

	/* set up command for data segment */
	ccmdp->ev_next = new_vmcmd(vmcmd_map_readvn,
	    execp->a_data,
	    epp->ep_daddr,
	    epp->ep_vp,
	    N_DATOFF(*execp, NMAGIC),
	    VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE);
	ccmdp = ccmdp->ev_next;

	/* set up command for bss segment */
	baddr = roundup(epp->ep_daddr + execp->a_data, NBPG);
	bsize = epp->ep_daddr + epp->ep_dsize - baddr;
	if (bsize > 0) {
		ccmdp->ev_next = new_vmcmd(vmcmd_map_zero, bsize, baddr,
		    0, 0, VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE);
		ccmdp = ccmdp->ev_next;
	}

	return exec_aout_setup_stack(p, epp, ccmdp);
}

/*
 * exec_aout_prep_omagic(): Prepare a 'native' OMAGIC binary's exec package
 */

int
exec_aout_prep_omagic(p, epp)
	struct proc *p;
	struct exec_package *epp;
{
	struct exec *execp = epp->ep_execp;
	struct exec_vmcmd *ccmdp;
	long bsize, baddr;

	epp->ep_taddr = N_TXTADDR(*execp, OMAGIC);
	epp->ep_tsize = execp->a_text;
	epp->ep_daddr = N_DATADDR(*execp, OMAGIC);
	epp->ep_dsize = execp->a_data + execp->a_bss;
	epp->ep_entry = execp->a_entry;

	/* set up command for text and data segments */
	epp->ep_vcp = new_vmcmd(vmcmd_map_readvn,
	    execp->a_text + execp->a_data,
	    epp->ep_taddr,
	    epp->ep_vp,
	    N_TXTOFF(*execp, OMAGIC),
	    VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE);
	ccmdp = epp->ep_vcp;

	/* set up command for bss segment */
	baddr = roundup(epp->ep_daddr + execp->a_data, __LDPGSZ);
	bsize = epp->ep_daddr + epp->ep_dsize - baddr;
	if (bsize > 0) {
		ccmdp->ev_next = new_vmcmd(vmcmd_map_zero, bsize, baddr,
		    0, 0, VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE);
		ccmdp = ccmdp->ev_next;
	}

	return exec_aout_setup_stack(p, epp, ccmdp);
}

/*
 * exec_aout_setup_stack(): Set up the stack segment for an a.out
 * executable.
 *
 * Note that the ep_ssize parameter must be set to be the current stack
 * limit; this is adjusted in the body of execve() to yield the
 * appropriate stack segment usage once the argument length is
 * calculated.
 */

int
exec_aout_setup_stack(p, epp, ccmdp)
	struct proc *p;
	struct exec_package *epp;
	struct exec_vmcmd *ccmdp;
{

	epp->ep_maxsaddr = USRSTACK - MAXSSIZ;
	epp->ep_minsaddr = USRSTACK;
	epp->ep_ssize = p->p_rlimit[RLIMIT_STACK].rlim_cur;

	/*
	 * set up commands for stack.  note that this takes *two*, one to
	 * map the part of the stack which we can access, and one to map
	 * the part which we can't.
	 *
	 * arguably, it could be made into one, but that would require the
	 * addition of another mapping proc, which is unnecessary
	 *
	 * note that in memory, things assumed to be: 0 ....... ep_maxsaddr
	 * <stack> ep_minsaddr
	 */
	ccmdp->ev_next = new_vmcmd(vmcmd_map_zero,
	    ((epp->ep_minsaddr - epp->ep_ssize) - epp->ep_maxsaddr),
	    epp->ep_maxsaddr,
	    0,
	    0,
	    VM_PROT_NONE);
	ccmdp = ccmdp->ev_next;
	ccmdp->ev_next = new_vmcmd(vmcmd_map_zero,
	    epp->ep_ssize,
	    (epp->ep_minsaddr - epp->ep_ssize),
	    0,
	    0,
	    VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE);

	return 0;
}
