/*-
 * Copyright (C) 1993, 1994 Christopher G. Demetriou
 * Copyright (C) 1992 Wolfgang Solfrank.
 * Copyright (C) 1992 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: kern_exec.c,v 1.40 1994/01/16 03:10:05 cgd Exp $
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/filedesc.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#include <sys/mount.h>
#include <sys/malloc.h>
#include <sys/namei.h>
#include <sys/vnode.h>
#include <sys/file.h>
#include <sys/acct.h>
#include <sys/exec.h>
#include <sys/ktrace.h>
#include <sys/resourcevar.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <sys/signalvar.h>

#include <vm/vm.h>
#include <vm/vm_kern.h>

#include <sys/kinfo_proc.h>

#include <machine/cpu.h>
#include <machine/reg.h>

#ifdef COPY_SIGCODE
extern char sigcode[], esigcode[];

#define	szsigcode	(esigcode - sigcode)
#else
#define	szsigcode	0
#endif

/*
 * check exec:
 * given an "executable" described in the exec package's namei info,
 * see what we can do with it.
 *
 * on entry, the exec_package is supposed to have properly initialized
 * values.  the nameidata pointed to by it must contain a valid segment
 * flag and a valid name; these are used to look up the executable to
 * be executed.
 *
 * on error exit, the vnode for the to-be-executed program has been
 * released, the namei pathname buffer has been deallocated, and
 * any argument lists or file descriptors held by the exec package
 * have been deallocated in the appropriate way.
 *
 * on successful exit, the program's vnode is return to the caller
 * locked, and any argument lists or file descriptors created by
 * the exec switch entry are returned to the caller unmolested.
 */
int
check_exec(p, epp)
	struct proc *p;
	struct exec_package *epp;
{
	int error, i;
	struct vnode *vp;
	char *cp, *ep, *name;
	struct nameidata *ndp = epp->ep_ndp;
	int resid;

	ndp->ni_nameiop = LOOKUP | FOLLOW | LOCKLEAF | SAVENAME;
	/* first get the vnode */
	if (error = namei(ndp, p)) {
		return error;
	}
	epp->ep_vp = vp = ndp->ni_vp;

	/* check for regular file */
	if (vp->v_type != VREG) {
		error = EACCES;
		goto bad1;
	}

	/* get attributes */
	if (error = VOP_GETATTR(vp, epp->ep_vap, p->p_ucred, p))
		goto bad1;

	/* Check mount point */
	if (vp->v_mount->mnt_flag & MNT_NOEXEC) {
		error = EACCES;
		goto bad1;
	}
	if ((vp->v_mount->mnt_flag & MNT_NOSUID) || (p->p_flag & STRC))
		epp->ep_vap->va_mode &= ~(VSUID | VSGID);

	/* check access.  for root we have to see if any exec bit on */
	if (error = VOP_ACCESS(vp, VEXEC, p->p_ucred, p))
		goto bad1;
	if (( epp->ep_vap->va_mode & 0111) == 0) {
		error = EACCES;
		goto bad1;
	}

	/* try to open it */
	if (error = VOP_OPEN(vp, FREAD, p->p_ucred, p))
		goto bad1;

	/* now we have the file, get the exec header */
	if (error = vn_rdwr(UIO_READ, vp, epp->ep_hdr, epp->ep_hdrlen, 0,
		UIO_SYSSPACE, IO_NODELOCKED, p->p_ucred, &resid, p))
		goto bad2;
	epp->ep_hdrvalid = epp->ep_hdrlen - resid;

	/*
	 * set up the vmcmds for creation of the process
	 * address space
	 */
	error = ENOEXEC;
	for (i = 0; i < nexecs && error == ENOEXEC; i++) {
		/*
		 * when calling exec switch function, vnode is
		 * locked.  It is expected to return containing
		 * a valid vnode and pathname buffer, even on
		 * error, because they're deallocated at the end of
		 * this routine.  If the function has allocated
		 * a file descriptor, it's assumed to be
		 * describing the vnode held in the exec package.
		 */
		if (execsw[i].es_check != NULL)
			error = (*execsw[i].es_check)(p, epp);
	}
	if (error)
		goto bad2;

	/* check that entry point is sane */
	if (epp->ep_entry > VM_MAXUSER_ADDRESS) {
		error = ENOEXEC;
		goto bad2;
	}

	/* check limits */
	if ((epp->ep_tsize > MAXTSIZ) ||
	    (epp->ep_dsize > p->p_rlimit[RLIMIT_DATA].rlim_cur)) {
		error = ENOMEM;
		goto bad2;
	}

	return 0;

bad2:
	/* kill any file descriptor or vnode held */
	if (epp->ep_flags & EXEC_HASFD) {
		exec_closefd(p, epp->ep_fd);
		VOP_UNLOCK(vp);
		epp->ep_flags &= ~EXEC_HASFD;
	} else {
		VOP_UNLOCK(vp);
		vn_close(vp, FREAD, p->p_ucred, p);
	}
	/* kill any argument lists held */
	if (epp->ep_flags & EXEC_HASARGL) {
		char **tmpfap = epp->ep_fa;
		while (*tmpfap != NULL) {
			FREE(*tmpfap, M_EXEC);
			tmpfap++;
		}
		FREE(epp->ep_fa, M_EXEC);
		epp->ep_flags &= ~EXEC_HASARGL;
	}
	/* free the namei pathname buffer */
	FREE(ndp->ni_pnbuf, M_NAMEI);
	return error;

bad1:
	/*
	 * free the namei pathname buffer, and put the vnode
	 * (which we don't yet have open).
	 */
	FREE(ndp->ni_pnbuf, M_NAMEI);
	vput(vp);
	return error;
}

/*
 * exec system call
 */
struct execve_args {
	char *fname;
	char **argp;
	char **envp;
};

/* ARGSUSED */
execve(p, uap, retval)
	register struct proc *p;
	register struct execve_args *uap;
	int *retval;
{
	int error, i;
	struct exec_package pack;
	struct nameidata nid;
	struct vattr attr;
	struct ucred *cred = p->p_ucred;
	char *argp;
	char **cpp, *dp, *sp, *np;
	int argc, envc, len;
	char *stack;
	struct ps_strings arginfo;
	struct vmspace *vm = p->p_vmspace;
	char **tmpfap;

	/* init the namei data to point the file user's program name */
	nid.ni_dirp = uap->fname;
	nid.ni_segflg = UIO_USERSPACE;

	/*
	 * initialize the fields of the exec package.
	 * XXX the header length shouldn't be initted to maxinterp,
	 * it should be allocated based on the maximum size needed by
	 * any exec switch entry.
	 */
	pack.ep_name = uap->fname;
	MALLOC(pack.ep_hdr, void *, MAXINTERP, M_EXEC, M_WAITOK); /* XXX */
	pack.ep_hdrlen = MAXINTERP;				  /* XXX */
	pack.ep_hdrvalid = 0;
	pack.ep_ndp = &nid;
	pack.ep_setup = NULL;	/* assume no setup function */
	pack.ep_vmcmds.evs_cnt = 0;
	pack.ep_vmcmds.evs_used = 0;
	pack.ep_vap = &attr;
	pack.ep_emul = EMUL_NETBSD;
	pack.ep_flags = 0;
	
	/* see if we can run it. */
	if (error = check_exec(p, &pack))
		goto freehdr;

	/* XXX -- THE FOLLOWING SECTION NEEDS MAJOR CLEANUP */

	/* allocate an argument buffer */
	argp = (char *) kmem_alloc_wait(exec_map, NCARGS);
#ifdef DIAGNOSTIC
	if (argp == (vm_offset_t) 0)
		panic("execve: argp == NULL");
#endif
	dp = argp;
	argc = 0;

	/* copy the fake args list, if there's one, freeing it as we go */
	if (pack.ep_flags & EXEC_HASARGL) {
		tmpfap = pack.ep_fa;
		while (*tmpfap != NULL) {
			char *cp;

			cp = *tmpfap;
			while (*cp)
				*dp++ = *cp++;
			*dp++ = 0;

			FREE(*tmpfap, M_EXEC);
			tmpfap++; argc++;
		}
		FREE(pack.ep_fa, M_EXEC);
		pack.ep_flags &= ~EXEC_HASARGL;
	}

	/* Now get argv & environment */
	if (!(cpp = uap->argp)) {
		error = EINVAL;
		goto bad;
	}

	if (pack.ep_flags & EXEC_SKIPARG)
		cpp++;

	while (1) {
		len = argp + ARG_MAX - dp;
		if (error = copyin(cpp, &sp, sizeof(sp)))
			goto bad;
		if (!sp)
			break;
		if (error = copyinstr(sp, dp, len, (u_int *) & len)) {
			if (error == ENAMETOOLONG)
				error = E2BIG;
			goto bad;
		}
		dp += len;
		cpp++;
		argc++;
	}

	envc = 0;
	if (cpp = uap->envp) {	/* environment need not be there */
		while (1) {
			len = argp + ARG_MAX - dp;
			if (error = copyin(cpp, &sp, sizeof(sp)))
				goto bad;
			if (!sp)
				break;
			if (error = copyinstr(sp, dp, len, (u_int *) & len)) {
				if (error == ENAMETOOLONG)
					error = E2BIG;
				goto bad;
			}
			dp += len;
			cpp++;
			envc++;
		}
		dp = (char *) ALIGN(dp);
	}

	/* Now check if args & environ fit into new stack */
	len = ((argc + envc + 2) * sizeof(char *) + sizeof(int) +
	    dp + STACKGAPLEN + szsigcode + sizeof(struct ps_strings)) - argp;
	len = ALIGN(len);	/* make the stack "safely" aligned */

	if (len > pack.ep_ssize) { /* in effect, compare to initial limit */
		error = ENOMEM;
		goto bad;
	}

	/* adjust "active stack depth" for process VSZ */
	pack.ep_ssize = len;	/* maybe should go elsewhere, but... */

	/* Unmap old program */
#ifdef sparc
	kill_user_windows(p);		/* before stack addresses go away */
#endif
	/* Kill shared memory and unmap old program */
#ifdef SYSVSHM
	if (vm->vm_shm)
		shmexit(p);
#endif
	vm_deallocate(&vm->vm_map, VM_MIN_ADDRESS,
		VM_MAXUSER_ADDRESS - VM_MIN_ADDRESS);

	/* Now map address space */
	vm->vm_taddr = (char *) pack.ep_taddr;
	vm->vm_tsize = btoc(pack.ep_tsize);
	vm->vm_daddr = (char *) pack.ep_daddr;
	vm->vm_dsize = btoc(pack.ep_dsize);
	vm->vm_ssize = btoc(pack.ep_ssize);
	vm->vm_maxsaddr = (char *) pack.ep_maxsaddr;

	/* create the new process's VM space by running the vmcmds */
#ifdef DIAGNOSTIC
	if (pack.ep_vmcmds.evs_used == 0)
		panic("execve: no vmcmds");
#endif
	for (i = 0; i < pack.ep_vmcmds.evs_used && !error; i++) {
		struct exec_vmcmd *vcp;

		vcp = &pack.ep_vmcmds.evs_cmds[i];
		error = (*vcp->ev_proc)(p, vcp);
	}

	/* free the vmspace-creation commands, and release their references */
	kill_vmcmds(&pack.ep_vmcmds);

	/* if an error happened, deallocate and punt */
	if (error)
		goto exec_abort;

	/* remember information about the process */
	arginfo.ps_nargvstr = argc;
	arginfo.ps_nenvstr = envc;

	/* Now copy argc, args & environ to new stack */
	stack = (char *) (USRSTACK - len);
	cpp = (char **) stack;

	if (copyout(&argc, cpp++, sizeof(argc)))
		goto exec_abort;
	dp = (char *) (cpp + argc + envc + 2);

	/* XXX don't copy them out, remap them! */
	arginfo.ps_argvstr = dp; /* remember location of argv for later */
	for (sp = argp; --argc >= 0; sp += len, dp += len) {
		len = strlen(sp) + 1;
		if (copyout(&dp, cpp++, sizeof(dp))
		    || copyoutstr(sp, dp, len, 0))
			goto exec_abort;
	}
	np = 0;
	if (copyout(&np, cpp++, sizeof(np)))
		goto exec_abort;

	arginfo.ps_envstr = dp;	/* remember location of env for later */
	for (; --envc >= 0; sp += len, dp += len) {
		len = strlen(sp) + 1;
		if (copyout(&dp, cpp++, sizeof(dp))
		    || copyoutstr(sp, dp, len, 0))
			goto exec_abort;
	}
	if (copyout(&np, cpp, sizeof(np)))
		goto exec_abort;

	/* copy out the process's ps_strings structure */
	if (copyout(&arginfo, (char *) PS_STRINGS, sizeof(arginfo)))
		goto exec_abort;

#ifdef COPY_SIGCODE
	/* copy out the process's signal trapoline code */
	if (copyout((char *) sigcode, ((char *) PS_STRINGS) - szsigcode,
		szsigcode)) {
		goto exec_abort;
	}
#endif

	fdcloseexec(p);		/* handle close on exec */
	execsigs(p);		/* reset catched signals */

	/* set command name & other accounting info */
	len = MIN(nid.ni_namelen, MAXCOMLEN);
	bcopy(nid.ni_ptr, p->p_comm, len);
	p->p_comm[len] = 0;
	p->p_acflag &= ~AFORK;

	/* record proc's vnode, for use by procfs and others */
        if (p->p_textvp)
                vrele(p->p_textvp);
	VREF(pack.ep_vp);
	p->p_textvp = pack.ep_vp;

	p->p_flag |= SEXEC;
	if (p->p_flag & SPPWAIT) {
		p->p_flag &= ~SPPWAIT;
		wakeup((caddr_t) p->p_pptr);
	}

	/*
	 * deal with set[ug]id.
	 * MNT_NOEXEC and STRC have already been used to disable s[ug]id.
	 */
	if (attr.va_mode & VSUID) {
		p->p_ucred = crcopy(cred);
		p->p_ucred->cr_uid = attr.va_uid;
	}
	p->p_cred->p_svuid = p->p_ucred->cr_uid;
	if (attr.va_mode & VSGID) {
		p->p_ucred = crcopy(p->p_ucred);
		p->p_ucred->cr_gid = attr.va_gid;
	}
	p->p_cred->p_svgid = p->p_ucred->cr_gid;

	kmem_free_wakeup(exec_map, (vm_offset_t) argp, NCARGS);

	FREE(nid.ni_pnbuf, M_NAMEI);
	VOP_CLOSE(pack.ep_vp, FREAD, cred, p);
	vput(pack.ep_vp);

	/* setup new registers and do misc. setup. */
	setregs(p, pack.ep_entry, (u_long) stack, retval);
	if (pack.ep_setup != NULL)
		 (*pack.ep_setup)(p, &pack);

	if (p->p_flag & STRC)
		psignal(p, SIGTRAP);

	p->p_emul = pack.ep_emul;
	FREE(pack.ep_hdr, M_EXEC);
	return EJUSTRETURN;

bad:
	if (pack.ep_flags & EXEC_HASFD) {
		pack.ep_flags &= ~EXEC_HASFD;
		exec_closefd(p, pack.ep_fd);
	}
	VOP_CLOSE(pack.ep_vp, FREAD, cred, p);
	vput(pack.ep_vp);
	FREE(nid.ni_pnbuf, M_NAMEI);
	kmem_free_wakeup(exec_map, (vm_offset_t) argp, NCARGS);
freehdr:
	FREE(pack.ep_hdr, M_EXEC);
	return error;

exec_abort:
	/*
	 * the old process doesn't exist anymore.  exit gracefully.
	 * get rid of the (new) address space we have created, if any, get rid
	 * of our namei data and vnode, and exit noting failure
	 */
	vm_deallocate(&vm->vm_map, VM_MIN_ADDRESS,
		VM_MAXUSER_ADDRESS - VM_MIN_ADDRESS);
	FREE(nid.ni_pnbuf, M_NAMEI);
	VOP_CLOSE(pack.ep_vp, FREAD, cred, p);
	vput(pack.ep_vp);
	kmem_free_wakeup(exec_map, (vm_offset_t) argp, NCARGS);
	FREE(pack.ep_hdr, M_EXEC);
	kexit(p, W_EXITCODE(0, SIGABRT));
	kexit(p, -1);

	/* NOTREACHED */
	return 0;
}
