/*
 * Copyright (c) 1993 Christopher G. Demetriou
 * Terms are the same as those of the CMU copyright below.
 *
 *	$Id: vfs_bio.c,v 1.13 1993/12/18 04:22:36 mycroft Exp $
 */

/*
 * Mach Operating System
 * Copyright (c) 1992 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon 
 * the rights to redistribute these changes.
 */

/*
 * Copyright (c) 1982, 1986, 1989 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	from: @(#)vfs_bio.c	7.40 (Berkeley) 5/8/91
 */

#include <sys/param.h>
#include <sys/proc.h>
#include <sys/buf.h>
#include <miscfs/specfs/specdev.h>
#include <sys/vnode.h>
#include <sys/mount.h>
#include <sys/trace.h>
#include <sys/resourcevar.h>

#include <vm/vm.h>
#include <vm/vm_param.h>

#ifdef BIO_DEBUG
int debug_bio = 0;
#define DPRINTF(x) if (debug_bio) printf x ;
#else
#define DPRINTF(x)
#endif

static struct buf	*getnewbuf  __P((void));
extern vm_map_t 	buffer_map;

struct buf	*buf;			/* the buffer pool itself */
char		*buffers;
int		nbuf;			/* number of buffer headers */
int		bufpages;	 /* number of memory pages in the buffer pool */
struct buf	*swbuf;		      /* swap I/O headers (XXX go elsewhere?) */
int		nswbuf;
struct bufhd	bufhash[BUFHSZ];	/* heads of hash lists */
struct buf	bfreelist[BQUEUES];	/* heads of available lists */
struct buf	bswlist;		/* head of free swap header list */
struct buf	*bclnlist;		/* head of cleaned page list */

/*
 * Initialize buffers and hash links for buffers.
 *
 * Ideas for initial buffer sizing (base & residual) taken from machdep.c
 */
void
bufinit()
{
	struct bufhd *bh;
	struct buf *bp;
	int base, pages, residual;
	int i;

	/* first, make a null hash table */
	for(bh = bufhash; bh < bufhash + BUFHSZ; bh++) {
		bh->b_flags = B_HEAD;
		bh->b_forw = (struct buf *)bh;
		bh->b_back = (struct buf *)bh;
	}

	/* next, make a null set of free lists */
	for(bp = bfreelist; bp < bfreelist + BQUEUES; bp++) {
		bp->b_flags = 0;
		bp->av_forw = bp;
		bp->av_back = bp;
		bp->b_forw = bp;
		bp->b_back = bp;
	}

	base = bufpages / nbuf;
	residual = bufpages % nbuf;

	/*
	 * initialize each buffer with the appropriate amount of space, and
	 * stick on the appropriate queue
	 */
	for (i = 0; i < nbuf; i++) {
		bp = &buf[i];
		pages = (i < residual ? base + 1 : base);

		bp->b_flags = B_INVAL;
		bp->b_dev = NODEV;
		bp->b_rcred = bp->b_wcred = NOCRED;
		bp->b_vp = NULL;
		bp->b_bufsize = CLBYTES * pages;
		bp->b_un.b_addr = buffers + (i * MAXBSIZE);

		if (pages == 0) {
			binstailfree(bp, bfreelist + BQ_EMPTY);
			binshash(bp, bfreelist + BQ_EMPTY);
		} else {
			binstailfree(bp, bfreelist + BQ_AGE);
			binshash(bp, bfreelist + BQ_AGE);
		}
		DPRINTF(("set up buffer; i = %d, bufsize = %x, addr = %x\n",
		    i, bp->b_bufsize, bp->b_un.b_addr));
	}
}


/*
 * Find the block in the buffer pool.
 * If the buffer is not present, allocate a new buffer and load
 * its contents according to the filesystem fill routine.
 */
int
bread(vp, blkno, size, cred, bpp)
	struct vnode *vp;
	daddr_t blkno;
	int size;
	struct ucred *cred;
	struct buf **bpp;
{
	struct buf *bp;
	int rv = 0;

	bp = getblk (vp, blkno, size);

	/* if not found in cache, do some I/O */
	if ((bp->b_flags & B_CACHE) == 0 || (bp->b_flags & B_INVAL) != 0) {
		bp->b_flags |= B_READ;
		bp->b_flags &= ~(B_DONE|B_ERROR|B_INVAL);
		if (cred != NOCRED && bp->b_rcred == NOCRED) {
			crhold(cred);
			bp->b_rcred = cred;
		}
		VOP_STRATEGY(bp);
		curproc->p_stats->p_ru.ru_inblock++;	/* pay for it */
		rv = biowait (bp);
	}
	*bpp = bp;
	
	return (rv);
}

/*
 * Operates like bread, but also starts I/O on the specified
 * read-ahead block. [See page 55 of Bach's Book]
 */
int
breada(vp, blkno, size, rablkno, rabsize, cred, bpp)
	struct vnode *vp;
	daddr_t blkno; 
	int size;
	daddr_t rablkno; 
	int rabsize;
	struct ucred *cred;
	struct buf **bpp;
{
	struct buf *bp, *rabp;
	int rv = 0, needwait = 0;

	bp = getblk (vp, blkno, size);

	/* if not found in cache, do some I/O */
	if ((bp->b_flags & B_CACHE) == 0 || (bp->b_flags & B_INVAL) != 0) {
		bp->b_flags |= B_READ;
		bp->b_flags &= ~(B_DONE|B_ERROR|B_INVAL);
		if (cred != NOCRED && bp->b_rcred == NOCRED) {
			crhold(cred);
			bp->b_rcred = cred;
		}
		VOP_STRATEGY(bp);
		curproc->p_stats->p_ru.ru_inblock++;	/* pay for it */
		needwait++;
	}
	
	rabp = getblk (vp, rablkno, rabsize);

	/* if not found in cache, do some I/O (overlapped with first) */
	if ((rabp->b_flags & B_CACHE) == 0 || (rabp->b_flags & B_INVAL) != 0) {
		rabp->b_flags |= B_READ | B_ASYNC;
		rabp->b_flags &= ~(B_DONE|B_ERROR|B_INVAL);
		if (cred != NOCRED && rabp->b_rcred  == NOCRED) {
			crhold(cred);
			rabp->b_rcred = cred;
		}
		VOP_STRATEGY(rabp);
		curproc->p_stats->p_ru.ru_inblock++; /* pay for it (up front) */
	} else
		brelse(rabp);
	
	/* wait for original I/O */
	if (needwait)
		rv = biowait (bp);

	*bpp = bp;
	return (rv);
}

/*
 * Synchronous write.
 * Release buffer on completion.
 */
int
bwrite(bp)
	register struct buf *bp;
{
	int rv;

	if(bp->b_flags & B_INVAL) {
		brelse(bp);
		return (0);
	} else {
		int wasdelayed;

		if(!(bp->b_flags & B_BUSY))
			panic("bwrite: not busy");

		wasdelayed = bp->b_flags & B_DELWRI;
		bp->b_flags &= ~(B_READ|B_DONE|B_ERROR|B_ASYNC|B_DELWRI);
		if(wasdelayed) 
			reassignbuf(bp, bp->b_vp);
		else
			curproc->p_stats->p_ru.ru_oublock++;	/* pay for it */

		bp->b_flags |= B_DIRTY;
		bp->b_vp->v_numoutput++;
		VOP_STRATEGY(bp);
		rv = biowait(bp);
		if (!rv)
			bp->b_flags &= ~B_DIRTY;
		brelse(bp);
		return (rv);
	}
}

/*
 * Delayed write.
 *
 * The buffer is marked dirty, but is not queued for I/O.
 * This routine should be used when the buffer is expected
 * to be modified again soon, typically a small write that
 * partially fills a buffer.
 *
 * NB: magnetic tapes cannot be delayed; they must be
 * written in the order that the writes are requested.
 */
void 
bdwrite(bp)
	register struct buf *bp;
{

	if(!(bp->b_flags & B_BUSY))
		panic("bdwrite: not busy");

	if(bp->b_flags & B_INVAL) {
		brelse(bp);
		return;
	}
	if(bp->b_flags & B_TAPE) {
		bwrite(bp);
		return;
	}
	bp->b_flags &= ~(B_READ|B_DONE);
	bp->b_flags |= B_DIRTY|B_DELWRI;
	reassignbuf(bp, bp->b_vp);
	curproc->p_stats->p_ru.ru_oublock++;	/* pay for it up front */
	brelse(bp);
	return;
}

/*
 * Asynchronous write.
 * Start I/O on a buffer, but do not wait for it to complete.
 * The buffer is released when the I/O completes.
 */
void
bawrite(bp)
	register struct buf *bp;
{

	if(!(bp->b_flags & B_BUSY))
		panic("bawrite: not busy");

	if(bp->b_flags & B_INVAL)
		brelse(bp);
	else {
		int wasdelayed;

		wasdelayed = bp->b_flags & B_DELWRI;
		bp->b_flags &= ~(B_READ|B_DONE|B_ERROR|B_DELWRI);
		if (wasdelayed)
			reassignbuf(bp, bp->b_vp);
		else
			curproc->p_stats->p_ru.ru_oublock++; /* pay for it */

		bp->b_flags |= B_DIRTY | B_ASYNC;
		bp->b_vp->v_numoutput++;
		VOP_STRATEGY(bp);
	}
}

/*
 * Release a buffer.
 * Even if the buffer is dirty, no I/O is started.
 */
void
brelse(bp)
	register struct buf *bp;
{
	int x;

	/* anyone need a "free" block? */
	x=splbio();
	if ((bfreelist + BQ_AGE)->b_flags & B_WANTED) {
		(bfreelist + BQ_AGE) ->b_flags &= ~B_WANTED;
		wakeup(bfreelist);
	}
	/* anyone need this very block? */
	if (bp->b_flags & B_WANTED) {
		bp->b_flags &= ~B_WANTED;
		wakeup(bp);
	}

	if (bp->b_flags & (B_INVAL|B_ERROR)) {
		bp->b_flags |= B_INVAL;
		bp->b_flags &= ~(B_DELWRI|B_CACHE);
		if(bp->b_vp)
			brelvp(bp);
	}

	/* enqueue */
	if (bp->b_bufsize == 0) {
		/* an empty buffer... */
		binsheadfree(bp, bfreelist + BQ_EMPTY)
	} else if (bp->b_flags & (B_ERROR|B_INVAL|B_NOCACHE)) {
		/* buffers with junk contents */
		binsheadfree(bp, bfreelist + BQ_AGE)
	} else if (bp->b_flags & B_AGE) {
		/* buffers with stale but valid contents */
		binstailfree(bp, bfreelist + BQ_AGE)
	} else {
		/* buffers with valid and quite potentially reuseable contents */
		binstailfree(bp, bfreelist + BQ_LRU)
	}

	/* unlock */
	bp->b_flags &= ~B_BUSY;
	splx(x);
}

/*
 * Find a buffer which is available for use.
 * Select something from a free list.
 * Preference is to AGE list, then LRU list.
 */
static struct buf *
getnewbuf()
{
	struct buf *bp;
	int x;

	x = splbio();

start:
	if (bfreelist[BQ_AGE].av_forw != (struct buf *)bfreelist+BQ_AGE) {
		bp = bfreelist[BQ_AGE].av_forw;
		bremfree(bp);
	} else if (bfreelist[BQ_LRU].av_forw != (struct buf *)bfreelist+BQ_LRU) {
		bp = bfreelist[BQ_LRU].av_forw;
		bremfree(bp);
	} else	{
		/* wait for a free buffer of any kind */
		(bfreelist + BQ_AGE)->b_flags |= B_WANTED;
		tsleep(bfreelist, PRIBIO, "getnewbuf", 0);
		splx(x);
		return (0);
	}

	/* if we are a delayed write, convert to an async write! */
	if (bp->b_flags & B_DELWRI) {
		bp->b_flags |= B_BUSY;
		bawrite (bp);
		goto start;
	}

	/* disassociate us from our vnode, if we had one... */
	if(bp->b_vp)
		brelvp(bp);

	/* we're a buffer that's on no free lists */
	bp->b_flags = B_BUSY;

	splx(x);

	/* clear out various fields */
	bp->b_dev = NODEV;
	bp->b_vp = NULL;
	bp->b_blkno = bp->b_lblkno = 0;
	bp->b_iodone = 0;
	bp->b_error = 0;
	bp->b_resid = 0;
	bp->b_bcount = 0;
	bp->b_dirtyoff = bp->b_dirtyend = 0;

	/* nuke any credentials we were holding */
	if (bp->b_rcred != NOCRED) {
		crfree(bp->b_rcred);
		bp->b_rcred = NOCRED;
	}
	if (bp->b_wcred != NOCRED) {
		crfree(bp->b_wcred);
		bp->b_wcred = NOCRED;
	}

	return (bp);
}

/*
 * Check to see if a block is currently memory resident.
 * This routine must be called at splbio().  If it is not,
 * and you take a disk interrupt while in the while loop,
 * you can loop forever.
 */
struct buf *
incore(vp, blkno)
	struct vnode *vp;
	daddr_t blkno;
{
	struct buf *bh;
	struct buf *bp;

	bh = BUFHASH(vp, blkno);

	/* Search hash chain */
	bp = bh->b_forw;
	while (bp != (struct buf *) bh) {
		/* hit */
		if (bp->b_lblkno == blkno && bp->b_vp == vp
			&& (bp->b_flags & B_INVAL) == 0)
			return (bp);
		bp = bp->b_forw;
	}
	
	return(0);
}

/*
 * Get a block of requested size that is associated with
 * a given vnode and block offset. If it is found in the
 * block cache, mark it as having been found, make it busy
 * and return it. Otherwise, return an empty block of the
 * correct size. It is up to the caller to insure that the
 * cached blocks be of the correct size.
 */
struct buf *
getblk(vp, blkno, size)
	register struct vnode *vp;
	daddr_t blkno;
	int size;
{
	struct buf *bp, *bh;
	int x;

	x = splbio();
start:
	if (bp = incore(vp, blkno)) {
		if (bp->b_flags & B_BUSY) {
			bp->b_flags |= B_WANTED;
			tsleep (bp, PRIBIO, "getblk", 0);
			goto start;
		}
		bp->b_flags |= B_BUSY | B_CACHE;
		bremfree(bp);
		allocbuf(bp, size);
	} else {
		splx(x);
		if((bp = getnewbuf()) == NULL) {
			x = splbio();
			goto start;
		}
		allocbuf(bp, size);
		bp->b_blkno = bp->b_lblkno = blkno;
		bgetvp(vp, bp);
		x = splbio();
		bremhash(bp);
		bh = BUFHASH(vp, blkno);
		binshash(bp, bh);
		bp->b_flags = B_BUSY;
	}
	splx(x);
	return (bp);
}

/*
 * Get an empty, disassociated buffer of given size.
 */
struct buf *
geteblk(size)
	int size;
{
	struct buf *bp;
	int x;

	while ((bp = getnewbuf()) == 0)
		;
	allocbuf(bp, size);
	x = splbio();
	bremhash(bp);
	binshash(bp, bfreelist + BQ_AGE);
	splx(x);

	return (bp);
}

/*
 * Expand or contract the actual memory allocated to a buffer.
 *
 * If the buffer shrinks, data is lost, so it's up to the
 * caller to have written it out *first*; this routine will not
 * start a write.  If the buffer grows, it's the callers
 * responsibility to fill out the buffer's additional contents.
 */
void
allocbuf(bp, size)
	register struct buf *bp;
	int size;
{
	struct buf	*nbp;
	vm_size_t	desired_size;
	int		s;

	desired_size = roundup(size, CLBYTES);
#ifdef DIAGNOSTIC
	if (desired_size == 0)
		panic("allocbuf: desired size is zero");
#endif
	if (desired_size > MAXBSIZE)
		panic("allocbuf: buffer larger than MAXBSIZE requested");

	if (bp->b_bufsize == desired_size)
		goto out;

	DPRINTF(("allocbuf (0x%x): have %d, want %d, so:", bp,
	    bp->b_bufsize, desired_size));

	/* If the buffer is smaller than the desired size, we need to snarf
	 * it from other buffers.  Get buffers (via getnewbuf()), and
	 * steal their pages.
	 */
	while (bp->b_bufsize < desired_size) {
		int amt;

		/* find a buffer */
		DPRINTF((" get"));
		while ((nbp = getnewbuf()) == NULL)
			;

		/* and steal its pages, up to the amount we need */
		amt = min(nbp->b_bufsize, (desired_size - bp->b_bufsize));
		DPRINTF((" and steal %d from 0x%x", amt, nbp));
		pagemove((nbp->b_un.b_addr + nbp->b_bufsize - amt),
			bp->b_un.b_addr + bp->b_bufsize, amt);
		bp->b_bufsize += amt;
		nbp->b_bufsize -= amt;
		DPRINTF((" (now has %d)\n", nbp->b_bufsize));

		/* reduce transfer count if we stole some data */
		if (nbp->b_bcount > nbp->b_bufsize)
			nbp->b_bcount = nbp->b_bufsize;

#ifdef BIO_DEBUG
		if (nbp->b_bufsize < 0)
			panic("allocbuf: negative bufsize");
#endif
		if (nbp->b_bufsize == 0) {
			bremhash(nbp);
			binshash(nbp, bfreelist + BQ_EMPTY);
			nbp->b_flags |= B_INVAL;
			nbp->b_error = 0;
			nbp->b_dev = NODEV;
		}
		brelse(nbp);
	}

	/*
	 * If we want a buffer smaller than the current size,
	 * shrink this buffer.  Grab a buf head from the EMPTY queue,
	 * move a page onto it, and put it on front of the AGE queue.
	 * If there are no free buffer headers, leave the buffer alone.
	 */
	if (bp->b_bufsize > desired_size) {
		if (bfreelist[BQ_EMPTY].av_forw ==
		    (struct buf *)&bfreelist[BQ_EMPTY]) {
			/* No free buffer head */
			DPRINTF((" fail to find empty buffer\n"));
			goto out;
		}

		/* we have our buffer... */
                nbp = bfreelist[BQ_EMPTY].av_forw;
		s = splbio();
                bremfree(nbp);
		nbp->b_flags |= B_BUSY;
		splx(s);

		/* move the page to it and note this change */
		DPRINTF((" give %d to 0x%x", bp->b_bufsize - desired_size, nbp));
		pagemove(bp->b_un.b_addr + desired_size,
		    nbp->b_un.b_addr, bp->b_bufsize - desired_size);
		nbp->b_bufsize = bp->b_bufsize - desired_size;
		bp->b_bufsize = desired_size;
		nbp->b_bcount = 0;
		nbp->b_flags |= B_INVAL;
		DPRINTF((" (now has %d)\n", nbp->b_bufsize));

		/* release the newly-filled buffer and leave */
		brelse(nbp);
	}

out:
	bp->b_bcount = size;
}

/*
 * Patiently await operations to complete on this buffer.
 * When they do, extract error value and return it.
 * Extract and return any errors associated with the I/O.
 * If an invalid block, force it off the lookup hash chains.
 */
int
biowait(bp)
	register struct buf *bp;
{
	int x;

	x = splbio();
	while ((bp->b_flags & B_DONE) == 0)
		tsleep((caddr_t)bp, PRIBIO, "biowait", 0);
	if((bp->b_flags & B_ERROR) || bp->b_error) {
		if ((bp->b_flags & B_INVAL) == 0) {
			bp->b_flags |= B_INVAL;
/* XXX
 * brelse() already puts buffers with B_ERROR set on the age queue
 *	and i believe it's wrong to let bp->b_error override B_ERROR
 *	but it certainly appears to work OK this way...
 */	
                      /* Some bufs aren't in the buffer cache. */
                      if (bp->b_back && bp->b_forw) {
                        bremhash(bp);
                        binshash(bp, bfreelist + BQ_AGE);
                      }
		}
		if (!bp->b_error)
			bp->b_error = EIO;
		else
			bp->b_flags |= B_ERROR;
		splx(x);
		return (bp->b_error);
	} else {
		splx(x);
		return (0);
	}
}

/*
 * Mark I/O complete on a buffer.
 *
 * If a callback has been requested, e.g. the pageout
 * daemon, do so. Otherwise, awaken waiting processes.
 *
 * [ daemon book says on p.274:
 *	"This routine wakes up the blocked process, frees the buffer
 *	 for an asynchronous write, or, for a request by the pagedaemon
 *	 process, invokes a procedure specified in the buffer structure" ]
 *
 *	In real life, the pagedaemon (or other system processes) wants
 *	to do async stuff to, and doesn't want the buffer brelse()'d.
 *	(for swap pager, that puts swap buffers on the free lists (!!!),
 *	for the vn device, that puts malloc'd buffers on the free lists!)
 */
void
biodone(bp)
	register struct buf *bp;
{
	bp->b_flags |= B_DONE;			/* note that it's done */

	if ((bp->b_flags & B_READ) == 0) 	/* wake up reader */
		vwakeup(bp);

	if (bp->b_flags & B_CALL) {		/* if necessary, call out */
		bp->b_flags &=  ~B_CALL;	/* but note callout done */
		(*bp->b_iodone)(bp);
	} else if (bp->b_flags & B_ASYNC) {	/* if async, release it */
		brelse(bp);
	} else {				/* or just wakeup the buffer */
		wakeup(bp);
                bp->b_flags &= ~B_WANTED;
	}
}

#ifdef BIO_DEBUG
void
bstats()
{
	struct buf *bp;
	int i, i2, sum;

	sum = 0;

	for (i2 = 0; i2 < BQUEUES; i2++) {
		i = 0; bp = bfreelist[i2].av_forw;
		while (bp != &bfreelist[i2]) {
			bp = bp->av_forw;
			i++;
		}
		printf("queue %d: %d\n", i2, i);
		sum += i;
	}

	printf("total: %d (%d)\n", sum, nbuf);
}

int whichlist;
void
bsizes()
{
	struct buf *bp;
	int i;

	printf("queue %d:\n", whichlist);
	i = 0; bp = bfreelist[whichlist].av_forw;
	while (bp != &bfreelist[whichlist]) {
		printf("buffer %d: at 0x%x, size %d\n", i, bp, bp->b_bufsize);
		bp = bp->av_forw;
		i++;
	}
}
#endif
